(*** This unit defines a class that implements a 3d Mesh and a class
implementing a list of Meshes.
The mesh can be loaded from an XML file with simple format.
This unit does not handle drawing of mesh, so it is not bound to any library.  
*)
unit meshunit;

(*

This unit was written by Nicola {MSX} Lugato.

http://msx.rules.it
nicola@lugato.net
nicola.lugato@gmail.com

It is released under the term of
The GNU General Public License as published by the Free
Software Foundation.

*)


interface

uses types3d, Nicola_xmlparser, sysutils, xml3d, classes, logger;

type
		(*** This class represent a Mesh, a collection of vertex and faces that forms a model. It has no
				 drawing implementation, so that it's not bound to any library. Extenders of this class can implement it. *)
    TMesh = class
     private
      nome:string;
      {** store radius of bounding sphere, so that it's calculated only once }
      radius: TFloat;                                           
      fNumFace, fNumVertex:Integer;                             
      procedure setNumface(value:integer);
      procedure setNumvertex(value:integer);
      procedure calculateBoundingSphereRadius;
     public
      userData:cardinal;                                        // use at your discretion
      Vertex: array of TVertex;                                 // dynamic array for vertex
      Face: array of TFace;                                     // dynamic array for faces
      {** used to recalculate normals. No so sure whether it works good :P }
      procedure CalcolateNormals;                               
      {** translate the mesh about its center (act on vertex! not on any view matrix) 
           @param(X,Y,Z the value of translation on the three axes)  }
      procedure Translate(x,y,z:TFloat);                        
      {** scale the mesh by a given factor (act on vertex! not on any view matrix) 
           @param(X,Y,Z the value of scaling on the single axes)  }
      procedure Scale(factor:TFloat);                         
      property NumFace:integer read fNumFace write SetNumFace;
      property NumVertex:integer read fNumVertex write SetNumVertex;
			{** load the mesh from an XML file }
      procedure LoadFrom(filename:String);overload;
      {** load the mesh from an XML stream }
      procedure LoadFrom(stream:TStream);overload; 
      {** return the name of the mesh }
      function getName():string;         
      constructor Create;virtual;                               
      {** Destroy the mesh releasing used resources }
      destructor Destroy;override;                     
      {** get the radius of the bounding sphere of this mesh }
      function getBoundingSphereRadius:TFloat;									
			{** This will be implemented by gfx-library specific code (opengl/directx/pen&paper/etc etc) }
			procedure Draw; virtual; abstract;                        
 
    end;

		 {** This class represent a list of meshes. You can obtain a mesh by index or by name }
     TMeshList=class(TList)
 			{** Return the mesh at specified index }
      function getMesh(index:integer):TMesh;
      property Mesh[index:integer]:TMesh read getMesh; default;
      {** Return the mesh with the given name or nil }
      function getMeshByName(name:string):TMesh;
     end;


implementation

function Tmesh.getName():string;
begin
result:=nome;
end;

procedure Tmesh.CalcolateNormals;
var i,i2,n:integer;
    s,x0,y0,z0,x1,y1,z1,x2,y2,z2,x,y,z:TFloat;
begin

// face normal

for i:=0 to NumFace-1 do
 begin
  x0:=vertex[face[i].point[1]].x;
  y0:=vertex[face[i].point[1]].y;
  z0:=vertex[face[i].point[1]].z;
  x1:=vertex[face[i].point[2]].x;
  y1:=vertex[face[i].point[2]].y;
  z1:=vertex[face[i].point[2]].z;
  x2:=vertex[face[i].point[3]].x;
  y2:=vertex[face[i].point[3]].y;
  z2:=vertex[face[i].point[3]].z;
  x:=(y1-y0)*(z2-z0)-(y2-y0)*(z1-z0);
  y:=(z1-z0)*(x2-x0)-(z2-z0)*(x1-x0);
  z:=(x1-x0)*(y2-y0)-(x2-x0)*(y1-y0);
  S:=(sqrt( x*x + y*y + z*z ));
  Face[i].nx:=x/S;
  Face[i].ny:=y/S;
  Face[i].nz:=z/S;
 end;

//vertex normal

for i:=0 to NumVertex-1 do
 begin
  n:=0; x:=0; y:=0; z:=0;
  for i2:=0 to numface-1 do
   if (Face[i2].point[1]=i) or
      (Face[i2].point[2]=i) or
      (Face[i2].point[3]=i) then
    begin
     inc(n);
     x:=x+face[i2].nx;
     y:=y+face[i2].ny;
     z:=z+face[i2].nz;
    end;
  Vertex[i].nx:=x/n;
  Vertex[i].ny:=y/n;
  Vertex[i].nz:=z/n;
 end;
end;

procedure TMesh.setNumface(value:integer);
begin
// quando cambiate NumFace, Tmodello setta
// la giusta dimensione dell'array Face
fNumFace:=value;
SetLength(face,fnumface);
end;

procedure TMesh.setNumvertex(value:integer);
begin
// quando cambiate NumVertex, Tmodello setta
// la giusta dimensione dell'array Vertex
fNumVertex:=value;
SetLength(vertex,fnumVertex);
end;

procedure TMesh.Scale(factor:TFloat);
var i:integer;
begin
// questa procedura scala un modello.
// con factor = 1 resta uguale
// con factor < 1 rimpicciolisce
// con factor > 1 ingrandisce
 for i:=0 to NumVertex-1 do
  begin
   Vertex[i].x:=Vertex[i].x*factor;
   Vertex[i].y:=Vertex[i].y*factor;
   Vertex[i].z:=Vertex[i].z*factor;
  end;
end;

procedure TMesh.Translate(x,y,z:TFloat);
var i:integer;
begin
 for i:=0 to NumVertex-1 do
  begin
   Vertex[i].x:=Vertex[i].x+x;
   Vertex[i].y:=Vertex[i].y+y;
   Vertex[i].z:=Vertex[i].z+z;
  end;
end;


procedure TMesh.LoadFrom(filename:String);
var s:TFileStream;
begin
log('Loading '+filename);
s:=TFileStream.Create(filename, fmopenread);
LoadFrom(s);
s.Free;
end;

procedure TMesh.LoadFrom(stream:TStream);
var
  i:integer;
  doc:TXMLDocument;
  vertici, faces, v:TXMLNode;
begin
doc:=TXMLDocument.Create;
doc.Load(stream);
nome:=doc.Node.Query('NAME').Text;

setNumface(doc.Node.Query('FACECOUNT').number);
setNumvertex(doc.Node.Query('POINTCOUNT').number);

vertici:=doc.Node.Query('POINTS');

for i:=0 to vertici.Children.Count-1 do
 begin
  if vertici.Children[i].Name='VERTEX' then
   begin
    v:=vertici.Children[i];
    vertex[strtoint(v.Attributes['id'])]:=XMLToVertex(v);
   end;
 end;

faces:=doc.Node.Query('FACES');

for i:=0 to faces.Children.Count-1 do
 begin
  if faces.Children[i].Name='FACE' then
   begin
    v:=faces.Children[i];
    face[strtoint(v.Attributes['id'])]:=XMLToFace(v);
   end;
 end;

doc.Free;
end;


function TMeshList.getMesh(index:integer):TMesh;
begin
result:=TMesh(items[index]);
end;

function TMeshList.getMeshByName(name:string):TMesh;
var i:integer;
begin
 for i:=0 to count-1 do
  if mesh[i].nome=name then
  begin
   result:=mesh[i];
   exit;
  end;
 result:=nil;
end;

constructor TMesh.Create;
begin
radius:=0;
userData:=0;
SetLength(vertex, 0);
SetLength(face, 0);
nome:='';
end;

destructor TMesh.Destroy;
begin
  SetLength(vertex, 0);
  SetLength(face, 0);
	inherited Destroy;
end;



procedure TMesh.calculateBoundingSphereRadius;
var i:integer;
    r:TFloat;
begin
 radius:=0;
 for i:=0 to NumVertex-1 do
  begin
   r := sqrt( sqr( Vertex[i].x ) + sqr( Vertex[i].y ) + sqr( Vertex[i].z ) );
   if radius<r then radius:=r;
  end;
end;
function TMesh.getBoundingSphereRadius:TFloat;
begin
if radius=0 then
  calculateBoundingSphereRadius;
result:=radius;
end;

end.
