{** 
This unit contains classes to handle models and animations.
A model is defined by a tree of Node that represent a hierarchy. Each node hold a graphical element (usually a mesh) and
a transformation to draw that element (rotation/position/scale) relatively to the parent node (plus other data). Each 
node has a name (used to reference it around). The tree definition represent the "static position" of the model.

A model also have a list of Poses. A pose is a different arrangement of the nodes. The pose is saved as a list of 
"moved nodes". For each moved node (ie a node that is in a different position from the "static" pose), an entry is saved 
with the node name and the new transformation (a diff from static transformation).

A model also have a list of animations. Each animation is a list of poses throu time. It is defined by a number of entries,
each having a pose name and a time value, to say that a given pose must be used at the given time. Time is from 0 to 99.
The animation will interpolate the "keys" inserted to obtain a smooth animation.


}
unit modelunit;

interface

uses
  meshunit, textureunit, msxmath, Contnrs, Classes, SysUtils, logger, 
  types3d, Nicola_xmlparser, xml3d, mediamanagerunit;

type

TModel = class;
TModelTreeNodeList = class;
TTransfTreeNode = class;

{** The possible types of graphical elements that can be attached to a tree node }
TGraphicElementType = (geNone, geMesh, geModel, geReference);
{** The graphical element attached to a tree node. }
TGraphicElement = record
 type_:TGraphicElementType;
 case TGraphicElementType of
   geNone: ();
   geMesh: (mesh:TMesh);
   geModel: (model:TModel);
   geReference: (reference:integer);
end;

{** Represent a transformation of the tree node. }
TTransformation = class
  // transformations (can this be done with a single matrix ?)
  Pos:TVector;
  Rot:TVector;
  Sca:TVector;
	
  // Piv:TVector; // is this useful?
  procedure setZero;
  function isZero:boolean;
  function mustToggleFaces:boolean;
  
  {** Return a new TTransformation that is the sum of self and x.
  The new object must be freed manually by the caller }
  function Sum(x:TTransformation):TTransformation;
  function Clone:TTransformation;
end;

{** The tree node that forms a model }
TModelTreeNode = class
  name:string;
  // node transformation
  transf:TTransformation;
  // mesh color (@todo use right data structure here)
  color:TRgb;
  // attacched item
  GraphicalElement:TGraphicElement;
  // texture (can be nil)
  texture:TTexture;
	
  // children of this tree node
  Children:TModelTreeNodeList;
 
  constructor Create;
  destructor Destroy;override;
  function load(xml:TXmlNode;man:TGenericMediaManager):boolean;
  function Save:TXmlNode;
  
  function CloneDeep:TModelTreeNode;
end;

TModelTreeNodeList = class(TObjectList)
  private
    function getNode(index:integer):TModelTreeNode;
  public
    property Node[index:integer]:TModelTreeNode read getNode; default;
end;

TTransfTreeNodeList = class(TObjectList)
  private
    function getTransfNode(index:integer):TTransfTreeNode;
  public
    property Node[index:integer]:TTransfTreeNode read getTransfNode; default;
end;


{** The tree node with only Transformations, useful for animations }
TTransfTreeNode = class
  // node transformation
  transf:TTransformation;
	
  // children of this tree node
  Children:TTransfTreeNodeList;
 
  constructor Create;
  destructor Destroy;override;

end;



{** This represent a node of the model tree that is moved in this pose (ie it is not in the position it is in the static model)
The node is referenced by name, and the moving is a diff from the original transformation }
TPoseItem = class
  movedNode:string;
  diff:TTransformation;
  constructor Create;
  destructor Destroy;override;
  function Save:TXmlNode;
  function load(xml:TXmlNode):boolean;
end;

TPoseItemList = class(TObjectList)
  private
    function getPoseItem(index:integer):TPoseItem;
  public
    property PoseItem[index:integer]:TPoseItem read getPoseItem; default;
end;

{** This represent a pose of the model (generally different from the static pose)
Each node of the model tree that is moved is saved within this class in a TPoseItem object. }
TPose = class
  name:string;
  Items:TPoseItemList;
  {** Get the (diff) TTransformation of the given node (by name). Eventually, if the node is not moved,
  it can create the (empty) transformation for the node }
  function getTransfForNode(nodename:string; createIfNotFound:boolean):TTransformation;
  {** Tells if the given node is moved relatively to the static position }
  function isMoved(nodename:string):boolean;
  constructor Create;
  destructor Destroy;override;
  function Save:TXmlNode;
  function load(xml:TXmlNode):boolean;

end;

TPoseList = class(TObjectList)
  private
    function getPose(index:integer):TPose;
  public
    property Pose[index:integer]:TPose read getPose; default;
end;


{** This is a key in an animation, that define what pose should be used
in a given time. Keys are interpolated in time }
TAnimationKey=class
 posename:string;
 time:integer;
 trigger:boolean;
 
 function Save:TXmlNode;
 function load(xml:TXmlNode):boolean;

end;

TAnimationKeyList = class(TObjectList)
  private
    function getAnimationKey(index:integer):TAnimationKey;
  public
    property AnimationKey[index:integer]:TAnimationKey read getAnimationKey; default;
end;


{** animation data of a model}
TAnimation = class
  name:string;
  keys:TAnimationKeyList;

  constructor Create;
  destructor Destroy;override;

  function Save:TXmlNode;
  function load(xml:TXmlNode):boolean;
end;

TAnimationList = class(TObjectList)
  private
    function getAnimation(index:integer):TAnimation;
  public
    property Animation[index:integer]:TAnimation read getAnimation; default;
end;


{** A 3d Model }
TModel = class
  name:string;

  Root:TModelTreeNode;
  
  Poses:TPoseList;
  Animations:TAnimationList;
  
  function Load(filename:string;man:TGenericMediaManager):boolean; overload;
  function Load(Stream:TStream;man:TGenericMediaManager):boolean; overload;
  function load(xml:TXmlNode;man:TGenericMediaManager):boolean;overload;
  function Save(filename:string):boolean; overload;
  function Save(Stream:TStream):boolean; overload;
  function Save:TXmlNode; overload;
  
  function hasPose(posename:string):boolean;
  function IndexOfPose(posename:string):integer;
  
  Constructor Create;
  
end;

TRealTimeAnimationKey = record pose, time:integer; end;
TRealTimeAnimation = record
 keys:array of TRealTimeAnimationKey;
 trigger:integer;
end;
TRealTimeAnimations = array of TRealTimeAnimation;

{** This represent a model with animation to be used in game. It's optimized for speed but not for editing.
You can load if from a TModel with build(Model). }
TRealTimeModel = class
  private
    fActualPose:TModelTreeNode;
    fPoseRoots:TTransfTreeNodeList;

    fAnimations:TRealTimeAnimations;

    {** Translate the pose of the model in a tree of transformation, summing static transf from model with diffs from pose }
    function CreateTreeForPose(pose:TPose; model:TModel):TTransfTreeNode;
    {** Recursively clone a node summing it with the pose to obtain a Transf Tree }
    function CloneTreeWithPose(pose:TPose; node:TModelTreeNode):TTransfTreeNode;
    procedure InterpolateTree(tree1, tree2:TTransfTreeNode; actual:TModelTreeNode; value:TFloat);
    procedure makeAnimation(var rta:TRealTimeAnimation; a:TAnimation; model:TModel);

    {** Fill fActualPose with the interpolation betwheen pose1 and pose2 at given position from 0 to 1 }
    procedure InterpolatePose(pose1, pose2:integer; value:TFloat);
  public
    procedure Build(Model:TModel);
    constructor Create;
    destructor Destroy;override;
    property Actual:TModelTreeNode read fActualPose;

    procedure Interpolate(Animation:integer; value:TFloat);
    function TriggerPassed(Animation:integer; value:TFloat):boolean;
    
    function Load(filename:string;man:TGenericMediaManager):boolean; 

  
end;



implementation


function TModelTreeNodeList.getNode(index:integer):TModelTreeNode;
begin
result:=TModelTreeNode(items[index]);
end;

constructor TModelTreeNode.Create;
begin
Children:=TModelTreeNodeList.create;
transf:=TTransformation.create;
GraphicalElement.type_:=geNone;
texture:=nil;

transf.Pos:=getVector(0,0,0);
transf.Rot:=getVector(0,0,0);
transf.Sca:=getVector(1,1,1);
color[0]:=1;
color[1]:=1;
color[2]:=1;

end;

destructor TModelTreeNode.Destroy;
begin
Children.free;

transf.free;
inherited destroy;
end;

function TModel.load(xml:TXmlNode;man:TGenericMediaManager):boolean;
var i:integer;
    c:Tpose;
    a:TAnimation;
begin
	name:=xml.query('NAME').text;
	Root.free;
	Root:=TModelTreeNode.create;
	result:=root.load(xml.query('NODE'),man);
 
 
  poses.clear;
  for i:=0 to xml.children.count-1 do
  if xml.children[i].Name='POSE' then
   begin
    c:=TPose.create;
    c.load(xml.children[i]);
    poses.add(c);
   end;

  Animations.clear;
  for i:=0 to xml.children.count-1 do
  if xml.children[i].Name='ANIMATION' then
   begin
    a:=TAnimation.create;
    a.load(xml.children[i]);
    Animations.add(a);
   end;

end;

function TModel.Save:TXmlNode;
var n:TXmlNode;
    i:integer;
begin
result:=TXmlNode.Create('MODEL');
result.AppendChildText('NAME', name );
result.Children.Add(Root.save);

for i:=0 to poses.Count-1 do
 result.Children.Add(poses[i].save);

for i:=0 to animations.Count-1 do
 result.Children.Add(animations[i].save);


end;

function TModel.Save(Stream:TStream):boolean;
var D:TXmlDocument;
begin
result:=false;
try
  d:=TXMLDocument.Create;
  d.Node:=Save;
  d.Save(stream);
  result:=true;
finally
  d.free;
end;
end;

function TModel.Load(Stream:TStream;man:TGenericMediaManager):boolean;
var D:TXmlDocument;
begin
result:=false;
try
  d:=TXMLDocument.Create;
  d.Load(Stream);
  load(d.Node,man);
  result:=true;
finally
  d.free;
end;
end;

function TModel.Save(filename:string):boolean;
var s:TFileStream;
begin
result:=false;
try
  s:=TFileStream.Create(filename, fmCreate);
  result:=Save(s);
finally
  s.free;
end;
end;


function TModel.Load(filename:string;man:TGenericMediaManager):boolean;
var s:TFileStream;
begin
result:=false;
if not FileExists(filename) then exit;
try
  s:=TFileStream.Create(filename, fmOpenRead);
  result:=Load(s,man);
finally
  s.free;
end;
end;

function TModelTreeNode.load(xml:TXmlNode;man:TGenericMediaManager):boolean;
var n:TXmlNode;
		t:string;
		i:integer;
		c:TModelTreeNode;
begin
	name:=xml.query('NAME').text;
  t:=xml.query('TEXTURE').text;
  if t='nil' then texture:=nil else texture:=man.getTextureByName(t);
	transf.pos := XMLTOVector(xml.query('POS'));
	transf.rot := XMLTOVector(xml.query('ROT'));
	transf.SCA := XMLTOVector(xml.query('SCA'));


	// colors
  n:=xml.query('COLOR');
  color[0]:=n.Query('RED').number/255;
  color[1]:=n.Query('GREEN').number/255;
  color[2]:=n.Query('BLUE').number/255;

	
	n:=xml.query('GRAPHICALELEMENT');
	t:=n.query('TYPE').Text;
	if t='none' then
	 begin
	 	 GraphicalElement.type_ := geNone;
	 end
	else if t='mesh' then
	 begin
	 	 GraphicalElement.type_ := geMesh;
     GraphicalElement.mesh:=man.getMeshByName(n.Query('MESH').text);
	 end
	else if t='model' then
	 begin
	 	 GraphicalElement.type_ := geModel;
	 end
	else if t='reference' then
	 begin
	 	 GraphicalElement.type_ := geReference;
	 end;
  
 for i:=0 to xml.children.count-1 do
  if xml.children[i].Name='NODE' then
   begin
   	 c:=TModelTreeNode.create;
   	 c.load(xml.children[i], man);
   	 children.add(c);
   end;

end;

function TModelTreeNode.Save:TXmlNode;
var i:integer;
    n:TXmlNode;
begin
result:=TXmlNode.create('NODE');
result.AppendChildText('NAME', name);
if texture <> nil then
  result.AppendChildText('TEXTURE', texture.name)
else
  result.AppendChildText('TEXTURE', 'nil');

n:=VectorToXML(transf.Pos);
n.Name:='POS';
result.Children.Add(n);

n:=VectorToXML(transf.Rot);
n.Name:='ROT';
result.Children.Add(n);

n:=VectorToXML(transf.Sca);
n.Name:='SCA';
result.Children.Add(n);

n:=TXmlNode.create('COLOR');
 n.AppendChildNumber('RED', trunc(color[0]*255));
 n.AppendChildNumber('GREEN', trunc(color[1]*255));
 n.AppendChildNumber('BLUE', trunc(color[2]*255));
result.Children.Add(n);

n:=TXmlNode.create('GRAPHICALELEMENT');

case GraphicalElement.type_ of
  geNone: 
    begin
    	n.AppendChildText('TYPE', 'none');
    end;
  geMesh: 
    begin
    	n.AppendChildText('TYPE', 'mesh');
    	n.AppendChildText('MESH', GraphicalElement.mesh.getname);
    end;
  geModel: 
    begin
    	n.AppendChildText('TYPE', 'model');
    	n.AppendChildText('MODEL', GraphicalElement.model.name);
    end;
  geReference: 
    begin
    	n.AppendChildText('TYPE', 'reference');
    	n.AppendChildNumber('REFERENCE', GraphicalElement.reference);
    end;
end;

result.Children.Add(n);

for i:=0 to Children.Count-1 do
 result.Children.Add(Children[i].save);
end;

procedure TTransformation.setZero;
begin
	Pos:=getVector(0,0,0);
	Rot:=getVector(0,0,0);
	Sca:=getVector(0,0,0);
end;

function TTransformation.isZero:boolean;
begin
result:=
         ( Pos.x=0 ) and ( Pos.y=0 ) and ( Pos.z=0 ) and
         ( Rot.x=0 ) and ( Rot.y=0 ) and ( Rot.z=0 ) and
         ( Sca.x=0 ) and ( Sca.y=0 ) and ( Sca.z=0 );

end;

function TTransformation.mustToggleFaces:boolean;
begin
result:=sca.x*sca.y*sca.z <0;
end;

function TTransformation.Sum(x:TTransformation):TTransformation;
begin
result:=TTransformation.create;
result.pos:=add(self.pos, x.pos);
result.rot:=add(self.rot, x.rot);
result.sca:=add(self.sca, x.sca);
end;

function TTransformation.Clone:TTransformation;
begin
result:=TTransformation.create;
result.pos:=self.pos;
result.rot:=self.rot;
result.sca:=self.sca;
end;




function TPose.getTransfForNode(nodename:string; createIfNotFound:boolean):TTransformation;
var i:integer;
    p:TPoseItem;
begin
 for i:=0 to Items.count-1 do
   begin
     if items[i].movedNode = nodename then
       begin
         result:=items[i].diff;
         exit;
       end;
   end;
  if createIfNotFound then
    begin
      p:=TPoseItem.Create;
      p.movedNode:=nodename;
      items.Add(p);
      p.diff.setZero;
      result:=p.diff;
    end
  else
    result:=nil;
end;

function TPose.isMoved(nodename:string):boolean;
var i:integer;
    p:TPoseItem;

begin
 for i:=0 to Items.count-1 do
   begin
     if items[i].movedNode = nodename then
       begin
         result:= not items[i].diff.isZero;
         exit;
       end;
   end;
 result:=false;
end;


function TPoseItemList.getPoseItem(index:integer):TPoseItem;
begin
result:=TPoseItem(items[index]);
end;

function TPoseList.getPose(index:integer):TPose;
begin
result:=TPose(items[index]);
end;


constructor TPoseItem.Create;
begin
diff:=TTransformation.create;
diff.setZero;
end;

destructor TPoseItem.Destroy;
begin
diff.free;
inherited Destroy;
end;

constructor TPose.Create;
begin
name:='A pose';
items:=TPoseItemList.create;
end;

destructor TPose.Destroy;
begin
items.free;
inherited Destroy;
end;

constructor TAnimation.Create;
begin
keys:=TAnimationKeyList.create;
name:='Animation';

end;

destructor TAnimation.Destroy;
begin
keys.free;
inherited Destroy;
end;



function TPoseItem.Save:TXmlNode;
var i:integer;
    n:TXmlNode;
begin
result:=TXmlNode.create('POSEITEM');
result.AppendChildText('MOVEDNODE', movedNode);

n:=VectorToXML(diff.Pos);
n.Name:='POS';
result.Children.Add(n);

n:=VectorToXML(diff.Rot);
n.Name:='ROT';
result.Children.Add(n);

n:=VectorToXML(diff.Sca);
n.Name:='SCA';
result.Children.Add(n);

end;


function TPoseItem.load(xml:TXmlNode):boolean;
var n:TXmlNode;
		t:string;
		i:integer;
		c:TModelTreeNode;
begin
  movedNode:=xml.query('MOVEDNODE').text;
  diff.pos := XMLTOVector(xml.query('POS'));
  diff.rot := XMLTOVector(xml.query('ROT'));
  diff.SCA := XMLTOVector(xml.query('SCA'));
end;

function TPose.Save:TXmlNode;
var i:integer;
begin
result:=TXmlNode.Create('POSE');
result.AppendChildText('NAME', name );

for i:=0 to items.Count-1 do
  if not items[i].diff.isZero then  // if is zero then no real movement so no need to save.
    result.Children.Add(items[i].save);

end;

function TPose.load(xml:TXmlNode):boolean;
var n:TXmlNode;
	  i:integer;
	  c:TPoseItem;
begin
	name:=xml.query('NAME').text;
	items.clear;
 for i:=0 to xml.children.count-1 do
  if xml.children[i].Name='POSEITEM' then
   begin
   	 c:=TPoseItem.create;
   	 c.load(xml.children[i]);
   	 items.add(c);
   end;
end;



function TAnimation.Save:TXmlNode;
var i:integer;
begin
result:=TXmlNode.Create('ANIMATION');
result.AppendChildText('NAME', name );

for i:=0 to keys.Count-1 do
  result.Children.Add(keys[i].save);

end;

function TAnimation.load(xml:TXmlNode):boolean;
var n:TXmlNode;
    i:integer;
    c:TAnimationKey;
begin
  name:=xml.query('NAME').text;
  keys.clear;
  for i:=0 to xml.children.count-1 do
  if xml.children[i].Name='ANIMITEM' then
   begin
    c:=TAnimationKey.create;
    c.load(xml.children[i]);
    keys.add(c);
   end;
end;

Constructor TModel.Create;
begin
name:='Amodel';
root:=TModelTreeNode.Create;
root.GraphicalElement.type_:=geNone;
root.texture:= nil;
root.name:='root';
Poses:=TPoseList.create;
Animations:=TAnimationList.create;
end;

function TAnimationlist.getAnimation(index:integer):TAnimation;
begin
result:=TAnimation(items[index]);
end;

function TAnimationKeylist.getAnimationKey(index:integer):TAnimationKey;
begin
result:=TAnimationKey(items[index]);
end;

function TAnimationKey.Save:TXmlNode;
var i:integer;
    n:TXmlNode;
begin
result:=TXmlNode.create('ANIMITEM');
result.AppendChildText('POSENAME', poseName );
result.AppendChildNumber('TIME', time);
if trigger then
 result.AppendChildText('TRIGGER', 'true' )
else
 result.AppendChildText('TRIGGER', 'false' )
end;


function TAnimationKey.load(xml:TXmlNode):boolean;
var n:TXmlNode;
begin
  posename:=xml.query('POSENAME').text;
  time:=xml.query('TIME').Number;
  n:=xml.query('TRIGGER');
  if n<>nil then 
    trigger:=lowercase(n.text)='true'
  else
    trigger:=false;
end;


function TModel.IndexOfPose(posename:string):integer;
var i:integer;
begin
for i:=0 to Poses.count-1 do
  if poses[i].name=posename then
    begin
     result:=i;
     exit;
    end;
result:=-1;
end;
function TModel.hasPose(posename:string):boolean;
var i:integer;
begin
for i:=0 to Poses.count-1 do
  if poses[i].name=posename then
    begin
     result:=true;
     exit;
    end;
result:=false;
end;


{** Recursively create the TTransfTreeNode by coping from node and eventually adding diffs from pose }
function TRealTimeModel.CloneTreeWithPose(pose:TPose; node:TModelTreeNode):TTransfTreeNode;
var poseTransf:TTransformation;
		i:integer;
begin
  // create the node
  result:=TTransfTreeNode.create;
  // free it's transf: we'll be creating it by ourself
  result.transf.free;
 
  if pose = nil then
  begin
    result.transf:=node.transf.clone;
  end
  else
  begin
    // let's see if this node is moved in this pose
    poseTransf:=pose.getTransfForNode(node.name, false);
    if poseTransf <> nil then
      // the node is  moved in the pose, so we sum pose diff to node transf
      result.transf:=node.transf.Sum(poseTransf)
    else
      // the node is not moved in this pose, so we take node transf.
      result.transf:=node.transf.clone;
  end;
  
  // recursively copy children
  for i:=0 to node.children.count-1 do
    result.children.add(CloneTreeWithPose(pose, node.children[i]));

end;


procedure TRealTimeModel.InterpolateTree(tree1, tree2:TTransfTreeNode; actual:TModelTreeNode; value:TFloat);
var i:integer;
begin
  msxmath.interpolate(tree1.transf.pos, tree2.transf.pos, value, actual.transf.pos);
  msxmath.interpolate(tree1.transf.rot, tree2.transf.rot, value, actual.transf.rot);
  msxmath.interpolate(tree1.transf.sca, tree2.transf.sca, value, actual.transf.sca);
  // recursively interpolate children
  for i:=0 to tree1.children.count-1 do
    InterpolateTree(tree1.children[i], tree2.children[i], actual.children[i], value);

end;

procedure TRealTimeModel.InterpolatePose(pose1, pose2:integer; value:TFloat);
begin
	InterpolateTree(fPoseRoots[pose1], fPoseRoots[pose2], fActualPose, value);
end;

function TRealTimeModel.CreateTreeForPose(pose:TPose; model:TModel):TTransfTreeNode;
begin
	result:=CloneTreeWithPose(pose, model.root);
end;

procedure TRealTimeModel.makeAnimation(var rta:TRealTimeAnimation; a:TAnimation; model:TModel);
var i:integer;
    sorted:boolean;
    aux:TRealTimeAnimationKey;
begin
rta.trigger := 50; // default trigger value, halfway the animation
setlength(rta.keys, a.keys.Count);
for i:=0 to a.Keys.Count-1 do
  begin
  	// store our trigger time (there should be only one key that is trigger)
  	if a.keys[i].trigger then rta.trigger:=a.keys[i].time;
  	
    rta.keys[i].pose:=model.IndexOfPose(a.Keys[i].posename);
    rta.keys[i].time:=a.Keys[i].time;
  end;

sorted:=false;
while not sorted do
  begin
    sorted:=true;
    for i:=0 to length(rta.keys)-2 do
     if rta.keys[i].time>rta.keys[i+1].time then
       begin
         sorted:=false;
         aux:=rta.keys[i];
         rta.keys[i]:=rta.keys[i+1];
         rta.keys[i+1]:=aux;
         break;
       end;
  end;

end;

procedure TRealTimeModel.Build(Model:TModel);
var i:integer;
    root:TTransfTreeNode;
begin
fPoseRoots.Clear;
// we add different "roots", one for each pose.
// in this way we have a complete tree for each pose
for i:=0 to model.Poses.count-1 do
  begin
    root:=CreateTreeForPose(model.poses[i], model);
    fPoseRoots.Add(root);
  end;
  
// the actual pose is initialized with the default model
fActualPose.free;
fActualPose:=model.Root.CloneDeep;

// build animations
setlength(fAnimations, model.Animations.Count);
for i:=0 to model.Animations.Count-1 do
  makeAnimation(fAnimations[i], model.Animations[i], model);
end;

constructor TRealTimeModel.Create;
begin
fPoseRoots:=TTransfTreeNodeList.create;
fActualpose:=nil;
end;

destructor TRealTimeModel.Destroy;
begin
fPoseRoots.free;
inherited;
end;


constructor TTransfTreeNode.Create;
begin
Children:=TTransfTreeNodeList.create;
transf:=TTransformation.create;
end;

destructor TTransfTreeNode.Destroy;
begin
Children.free;
transf.free;
inherited destroy;
end;


function TTransfTreeNodeList.getTransfNode(index:integer):TTransfTreeNode;
begin
result:=TTransfTreeNode(items[index]);
end;

procedure TRealTimeModel.Interpolate(Animation:integer; value:TFloat);
var i,p:integer;
    val:integer;
    d, v:TFloat;
    a:TRealTimeAnimation;
begin
if animation>=length(fanimations) then exit;

a:=fanimations[animation];
if length(a.keys) <= 1 then exit; // 1 or 0 poses means no interpolation!

val:=trunc(value*100);

// let's find which two poses are inside the "value" time
p:=length(a.keys)-1;  //this is the second animationitem index
for i:=1 to length(a.keys)-1 do
  if a.keys[i].time>val then
    begin
     p:=i;
     break;
    end;

// ok, poses are a[p-1] and a[p]

// this transform global animation time to pose-to-pose time
d:=(a.keys[p].time - a.keys[p-1].time)/100; // delta time
v:=value- (a.keys[p-1].time/100);
v:=v/d;

// finally, interpolate the poses
InterpolatePose(a.keys[p-1].pose, a.keys[p].pose, v);
end;

function TModelTreeNode.CloneDeep:TModelTreeNode;
var i:integer;
begin
result:=TModelTreeNode.create;
result.transf.free;
result.transf:=self.transf.Clone;
result.color:=self.color;
result.GraphicalElement:=self.GraphicalElement;
for i:=0 to Children.Count-1 do
 begin
   result.Children.Add(children[i].CloneDeep);
 end;
end;

function TRealTimeModel.TriggerPassed(Animation:integer; value:TFloat):boolean;
var a:TRealTimeAnimation;
begin
	a:=fanimations[animation];
	result:=(a.trigger/100)<=value;
end;

function TRealTimeModel.Load(filename:string;man:TGenericMediaManager):boolean;
var m:TModel;
begin
	log('Loading '+filename);
	m:=TModel.create;
	m.load(filename, man);
	build(m);
	m.free;
end;

end.

